/*-
 * Copyright (C)2008 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2008 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: argus1tocopy.c,v 1.4 2013/10/23 19:03:46 babolo Exp $";
#endif /* not lint */

#define MIFE_COMPAT3 1
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include "argus1tocopy.h"

#define ARGUS1_BS 60

static void
usage(int ex) {
    fprintf( stderr
           , "argus1tocopy @BABOLO V.M "VERS"  "DATE"\n"
             "   Usage:\n"
             "argus1tocopy [... ] itraf iaddr iscan\n"
             " -r infile\n"
             " -c polfile\n"
             " -m maclistfile\n"
             " -p pluspacket (default 42)\n"
             " -L size - traf cashe (default 16)\n"
             " -M size - addr cashe (default 16)\n"
             " -N size - scan cashe (default 16)\n"
             " -T bits - mask length for time (default 0)\n"
             " -v[v[v..]] - verbose\n"
             " -q - quiet\n"
           );
    exit(ex);
}

static argus1_policy defaultpolicy = {0, 0, 0, 1};

int
main(int argc, char **argv) {
    mife_descriptor *fin = NULL, *tmf;
    argus1tocopy *opts;
    int ex = EX_OK, c;
    ssize_t l, polen = 1;
    off_t o;
#   define SET(T, B, L, I, N, E) {                                                                    \
        int c;                                                                                        \
        if  (!(opts->B = malloc(opts->L * sizeof(T *)))) {                                            \
            ifBLIN_QV1(opts->flags) warnx("No RAM for buffer %d p" E, opts->L);                       \
            errno = ENOMEM;                                                                           \
            ex = EX_UNAVAILABLE;                                                                      \
            goto out;                                                                                 \
        }                                                                                             \
        if  (!(opts->B[0] = malloc(opts->L * sizeof(T)))) {                                           \
            ifBLIN_QV1(opts->flags) warnx("No RAM for %d b" E, opts->L);                              \
            errno = ENOMEM;                                                                           \
            ex = EX_UNAVAILABLE;                                                                      \
            goto out;                                                                                 \
        }                                                                                             \
        for (c = 1; c < opts->L; c++) opts->B[c] = &(opts->B[0][c]);                                  \
        if  (argc > N && argv && argv[N] && argv[N][0] && !(opts->I = fopen(argv[N], "w"))) {         \
            ifBLIN_QV1(opts->flags) warn("Cant\'t create %s", argv[N]);                               \
            ex = EX_CANTCREAT;                                                                        \
            goto out;                                                                                 \
    }   }

    if  (!(opts = malloc(sizeof(argus1tocopy)))) {
        ifBLIN_QV1(opts->flags) warn("No RAM #1");
        ex = EX_UNAVAILABLE;
        goto out;
    }
    bzero(opts, sizeof(argus1tocopy));
    opts->pluspacket = 42;
    opts->ltraf = 16;
    opts->laddr = 16;
    opts->lscan = 16;
    opts->flags = BLIN_VER1;
    while ((c = getopt(argc, argv, "c:hm:L:M:N:p:qr:T:v?")) > 0) {
        switch (c) {
        case 'c':
            if  (opts->policy) {
                ifBLIN_QV1(opts->flags) warnx("Another policy: %s", optarg);
                ex = EX_USAGE;
                errno = EINVAL;
                goto out;
            }
            if  (!(tmf = mife_open(0, optarg))) {
                ifBLIN_QV1(opts->flags) warn("Can't open policy: %s", optarg);
                ex = EX_NOINPUT;
                goto out;
            }
            if  ((polen = mife_read(tmf, 0, 0)) < 0) {
                ifBLIN_QV1(opts->flags) warn("Can't read policy: %s", optarg);
                ex = EX_NOINPUT;
                goto out;
            }
            if  (!(opts->policy = mife_get(tmf, 0))) {
                ifBLIN_QV1(opts->flags) warn("Can't get policy: %s", optarg);
                ex = EX_NOINPUT;
                goto out;
            }
            break;
        case 'm':
            if  (argus1_MAClist(opts, optarg) < 0) {
                ifBLIN_QV1(opts->flags) warn("MAC list error: %s", optarg);
                ex = EX_DATAERR;
                goto out;
            }
            break;
        case 'L':
            opts->ltraf = strtol(optarg, NULL, 0);
            break;
        case 'M':
            opts->laddr = strtol(optarg, NULL, 0);
            break;
        case 'N':
            opts->lscan = strtol(optarg, NULL, 0);
            break;
        case 'p':
            opts->pluspacket = strtol(optarg, NULL, 0);
            break;
        case 'q':
            opts->flags &= ~BLIN_VERMASK;
            break;
        case 'r':
            if  (fin) {
                ifBLIN_QV1(opts->flags) warnx("Another input: %s", optarg);
                ex = EX_USAGE;
                errno = EINVAL;
                goto out;
            }
            if  (!(fin = mife_open(0, optarg))) {
                ifBLIN_QV1(opts->flags) warn("Can't open: %s", optarg);
                ex = EX_NOINPUT;
                goto out;
            }
            break;
        case 'T':
            opts->timask = ~(~0 << strtol(optarg, NULL, 0));
            break;
        case 'v':
            BLIN_VERBOSE(opts->flags);
            break;
        case 'h': usage(EX_OK);
        default: usage(EX_USAGE);
    }   }

    argc -= optind;
    argv += optind;
    if  (!opts->policy) opts->policy = &defaultpolicy;
    if  (!fin && !(fin = mife_opef(0, fileno(stdin)))) {
        ifBLIN_QV1(opts->flags) warn("Can't open stdin");
        ex = EX_NOINPUT;
        goto out;
    }
    SET(argus1_traf, btraf, ltraf, itraf, 0, "traf")
    SET(argus1_addr, baddr, laddr, iaddr, 1, "addr")
    SET(argus1_scan, bscan, lscan, iscan, 2, "scan")
    ifBLIN_QV3(opts->flags) {
        fprintf( stderr, "flags=%08X pluspacket=%d ltraf=%d laddr=%d lscan=%d nmacs=%d:\n"
               , opts->flags, opts->pluspacket, opts->ltraf, opts->laddr, opts->lscan, opts->nmacs
               );
        for (c = 0; c < opts->nmacs; c++)
            fprintf( stderr, "%02X:%02X:%02X:%02X:%02X:%02X\n"
                   , (u_int32_t)(opts->maclist[0] >> 40) & 0xFF
                   , (u_int32_t)(opts->maclist[0] >> 32) & 0xFF
                   , (u_int32_t)(opts->maclist[0] >> 24) & 0xFF
                   , (u_int32_t)(opts->maclist[0] >> 16) & 0xFF
                   , (u_int32_t)(opts->maclist[0] >> 8) & 0xFF
                   , (u_int32_t)opts->maclist[0] & 0xFF
                   );
        fprintf( stderr, "%s%s%spolicy %d:"
               , opts->itraf ? "itraf " : ""
               , opts->iaddr ? "iaddr " : ""
               , opts->iscan ? "iscan " : ""
               , polen / sizeof(argus1_policy)
               );
        if  (polen % sizeof(argus1_policy)) {
            fprintf(stderr, " illegal reminder %d\n", polen % sizeof(argus1_policy));
        } else {
            fprintf(stderr, "\n");
        }
        for (c = 0; c < (polen / sizeof(argus1_policy)); c++) argus1_polout(opts, c);
    }
    for (o = 0; (l = mife_read(fin, ARGUS1_BS, o)) >= ARGUS1_BS; o += ARGUS1_BS) {
        ifBLIN_QV3(opts->flags)
            fprintf( stderr, "N %lld @ %lld %d<%d> %d<%d> %d<%d>\n"
                   , o / ARGUS1_BS, o
                   , opts->ptraf, opts->ltraf
                   , opts->paddr, opts->laddr
                   , opts->pscan, opts->lscan
                   );
        if  (argus1_decode(opts, mife_get(fin, o)) < 0) {
            ifBLIN_QV1(opts->flags) warnx("Argus record %lld decode failed", o / ARGUS1_BS);
            ex = EX_DATAERR;
            goto out;
        }
        while (opts->ptraf + 1 >= opts->ltraf) argus1_trafout(opts);
        while (opts->paddr + 1 >= opts->laddr) argus1_addrout(opts);
        while (opts->pscan + 1 >= opts->lscan) argus1_scanout(opts);
    }
    while (opts->ptraf) argus1_trafout(opts);
    while (opts->paddr) argus1_addrout(opts);
    while (opts->pscan) argus1_scanout(opts);
    if  (opts->itraf) fclose(opts->itraf);
    if  (opts->iaddr) fclose(opts->iaddr);
    if  (opts->iscan) fclose(opts->iscan);
out:
    exit(ex);
}
