/*-
 * Copyright (C)2008 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2008 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: maclist.c,v 1.3 2013/10/23 19:03:46 babolo Exp $";
#endif /* not lint */

#define MIFE_COMPAT3 1
#include <sys/types.h>
#include <net/ethernet.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include "argus1tocopy.h"

#define MACL_DELTA 512

#define L0 0 /* \0           */
#define Ll 1 /*   */
#define Lb 2 /*        */
#define Lc 3 /* #            */
#define Li 4 /* :            */
#define Ld 5 /* 0-9          */
#define Lu 6 /* A-F          */
#define Lh 7 /* a-f          */
#define Lo 8 /*     */
#define Lx 9

static u_char *clasnames[] =
{ "L0", "Ll", "Lb", "Lc", "Li", "Ld", "Lu", "Lh", "Lo", "Lx"};

static u_char class[256] =
{ L0, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lb, Ll, Ll,  Lo, Ll, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lb, Lo, Lo, Lc,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Li, Li, Lo
, Ld, Ld, Ld, Ld,  Ld, Ld, Ld, Ld,  Ld, Ld, Li, Lo,  Lo, Lo, Lo, Lo

, Lo, Lu, Lu, Lu,  Lu, Lu, Lu, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lh, Lh, Lh,  Lh, Lh, Lh, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo

, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo

, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
};

static const u_char *statenames[] =
{ "bg", "bd", "bi", "be", "bx"};

enum states
{ bg /*     */
, bd /*  */
, bi /*   */
, be /*     */
, bx
};
static const u_char *pn[] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "Cd", "Cu", "Ch", "Cn", "Cz", "- ", "- ", "Cx"
};

#define Cd  0x008000 /*                */
#define Cu  0x004000 /*            */
#define Ch  0x002000 /*             */
#define Cn  0x001000 /*         */
#define Cz  0x000800 /*           */
#define Cx  0x000100 /*               */
#define C_state 0xFF /*   */

static u_int32_t automa[Lx][bx] =
/*    bg     bd     bi  be */
{{    bx, Cz|bx, Cx|bx, bx}/* L0 */
,{    bg, Cz|bg, Cx|bx, bg}/* Ll */
,{    bg, Cz|be, Cx|bx, be}/* Lb */
,{    be, Cz|be, Cx|bx, be}/* Lc */
,{    bx, Cn|bi, Cx|bx, be}/* Li */
,{ Cd|bd, Cd|bd, Cd|bd, be}/* Ld */
,{ Cu|bd, Cu|bd, Cu|bd, be}/* Lu */
,{ Ch|bd, Ch|bd, Ch|bd, be}/* Lh */
,{    bx, Cx|bx, Cx|bx, be}/* Lo */
};

int
argus1_MAClist(argus1tocopy *opts, char *macfile) {
    int ex = EX_OK;
    u_int64_t r_mac, r_part;
    int r_lm, r_lp;
    struct stat sb;
    mife_descriptor *macf;
    size_t listl, offset;
    ssize_t textl;
    u_int32_t control;
    enum states state;
    u_char s, *text;
    int p, lfcount;

    opts->nmacs = 0;
    if  (!(macf = mife_open(0, macfile))) {
        ifBLIN_QV1(opts->flags) warn("MAC can't open: %s", macfile);
        ex = EX_NOINPUT;
        goto out;
    }
    if  (fstat(macf->f.fd, &sb) < 0) {
        ifBLIN_QV1(opts->flags) warn("MAC can't stat: %s", macfile);
        ex = EX_NOINPUT;
        goto out;
    }
    if  (S_ISREG(sb.st_mode)) {
        listl = sb.st_size / 12;
    } else {
        listl = MACL_DELTA;
    }
    if  ((textl = mife_read(macf, 0, 0)) < 0) {
        ifBLIN_QV1(opts->flags) warn("MAC can't read: %s", macfile);
        ex = EX_IOERR;
        goto out;
    }
    if  (!(text = mife_get(macf, 0))) {
        ifBLIN_QV1(opts->flags) warn("MAC can't get: %s", macfile);
        ex = EX_IOERR;
        goto out;
    }
    if  (!(opts->maclist = malloc(sizeof(u_int64_t) * listl))) {
        ifBLIN_QV1(opts->flags) warnx("No RAM for MAC list %d long", listl);
        errno = ENOMEM;
        ex = EX_UNAVAILABLE;
        goto out;
    }
    r_mac = r_part = 0;
    r_lm = r_lp = 0;
    lfcount = 1;
    ifBLIN_QV5(opts->flags) BLIN_STATEHEADER(stderr, pn)
    for (state = bg, offset = 0, control = 0; state < bx; offset++, state = control & C_state) {
        if  (offset < textl) {
            s = text[offset];
            p = class[s];
            if  (s == '\n') lfcount++;
        } else {
            p = L0;
            s = 0;
        }
        control = automa[p][state];
        ifBLIN_QV5(opts->flags)
            BLIN_STATEBODY(stderr, pn, (u_int32_t)r_lm & 0xFF, clasnames[p], text, offset, (u_int32_t)r_mac, control, statenames, state, C_state, (text[offset + BLIN_STATEVAR_i]>=' '))
        if  (control & Cd) r_part = (r_part << 4) | ((s - '0') & 0x0F);
        if  (control & Cu) r_part = (r_part << 4) | ((10 + (s - 'A')) & 0x0F);
        if  (control & Ch) r_part = (r_part << 4) | ((10 + (s - 'a')) & 0x0F);
        if  (control & (Cd | Cu | Ch)) r_lp++;
        if  (control & (Cn | Cz)) {
            r_lp = (r_lp + 1) & ~1;
            r_mac = (r_mac << (4 * r_lp)) | r_part;
            r_lm += r_lp;
            r_lp = 0;
        }
        if  (control & Cz) {
            if  (r_lm != ETHER_ADDR_LEN * 2) {
                ifBLIN_QV1(opts->flags) warnx("MAC syntax error in line %d", lfcount);
                errno = EINVAL;
                ex = EX_DATAERR;
                goto out;
            }
            if  (opts->nmacs >= listl)  {
                listl += MACL_DELTA;
                if  (!(opts->maclist = reallocf(opts->maclist, listl * sizeof(u_int64_t)))) {
                    ifBLIN_QV1(opts->flags) warnx("No RAM for MAC list %d long", listl);
                    errno = ENOMEM;
                    ex = EX_UNAVAILABLE;
                    goto out;
            }   }
            opts->maclist[opts->nmacs++] = r_mac;
        }
        if  (control & Cx) {
            ifBLIN_QV1(opts->flags) fprintf(stderr, "   MAC\n");
            errno = EINVAL;
            ex = EX_DATAERR;
    }   }
out:
    return(ex);
}
